@@ -1,8 +1,8 @@
 /* abandon.c */
-/* $OpenLDAP: pkg/ldap/libraries/libldap/abandon.c,v 1.41.2.4 2007/02/13 20:22:10 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2021 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -68,21 +68,17 @@ ldap_abandon_ext(
 {
 	int	rc;
 
-	Debug( LDAP_DEBUG_TRACE, "ldap_abandon_ext %d\n", msgid, 0, 0 );
+	Debug1( LDAP_DEBUG_TRACE, "ldap_abandon_ext %d\n", msgid );
 
 	/* check client controls */
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_lock( &ld->ld_req_mutex );
-#endif
+	LDAP_MUTEX_LOCK( &ld->ld_req_mutex );
 
 	rc = ldap_int_client_controls( ld, cctrls );
 	if ( rc == LDAP_SUCCESS ) {
 		rc = do_abandon( ld, msgid, msgid, sctrls, 1 );
 	}
 
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_unlock( &ld->ld_req_mutex );
-#endif
+	LDAP_MUTEX_UNLOCK( &ld->ld_req_mutex );
 
 	return rc;
 }
@@ -102,7 +98,7 @@ ldap_abandon_ext(
 int
 ldap_abandon( LDAP *ld, int msgid )
 {
-	Debug( LDAP_DEBUG_TRACE, "ldap_abandon %d\n", msgid, 0, 0 );
+	Debug1( LDAP_DEBUG_TRACE, "ldap_abandon %d\n", msgid );
 	return ldap_abandon_ext( ld, msgid, NULL, NULL ) == LDAP_SUCCESS
 		? 0 : -1;
 }
@@ -115,16 +111,9 @@ ldap_pvt_discard(
 {
 	int	rc;
 
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_lock( &ld->ld_req_mutex );
-#endif
-
+	LDAP_MUTEX_LOCK( &ld->ld_req_mutex );
 	rc = do_abandon( ld, msgid, msgid, NULL, 0 );
-
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_unlock( &ld->ld_req_mutex );
-#endif
-
+	LDAP_MUTEX_UNLOCK( &ld->ld_req_mutex );
 	return rc;
 }
 
@@ -141,8 +130,8 @@ do_abandon(
 	Sockbuf		*sb;
 	LDAPRequest	*lr;
 
-	Debug( LDAP_DEBUG_TRACE, "do_abandon origid %d, msgid %d\n",
-		origid, msgid, 0 );
+	Debug2( LDAP_DEBUG_TRACE, "do_abandon origid %d, msgid %d\n",
+		origid, msgid );
 
 	/* find the request that we are abandoning */
 start_again:;
@@ -180,13 +169,9 @@ start_again:;
 	/* ldap_msgdelete locks the res_mutex. Give up the req_mutex
 	 * while we're in there.
 	 */
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_unlock( &ld->ld_req_mutex );
-#endif
+	LDAP_MUTEX_UNLOCK( &ld->ld_req_mutex );
 	err = ldap_msgdelete( ld, msgid );
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_lock( &ld->ld_req_mutex );
-#endif
+	LDAP_MUTEX_LOCK( &ld->ld_req_mutex );
 	if ( err == 0 ) {
 		ld->ld_errno = LDAP_SUCCESS;
 		return LDAP_SUCCESS;
@@ -221,20 +206,24 @@ start_again:;
 			 *		LDAP_NEXT_MSGID(ld, i);
 			 */
 
-			i = ++(ld)->ld_msgid;
+			LDAP_NEXT_MSGID(ld, i);
 #ifdef LDAP_CONNECTIONLESS
 			if ( LDAP_IS_UDP(ld) ) {
-				err = ber_write( ber, ld->ld_options.ldo_peer,
-					sizeof(struct sockaddr), 0);
+				struct sockaddr_storage sa = {0};
+				/* dummy, filled with ldo_peer in request.c */
+				err = ber_write( ber, (char *) &sa, sizeof(sa), 0 );
 			}
 			if ( LDAP_IS_UDP(ld) && ld->ld_options.ldo_version ==
 				LDAP_VERSION2 )
 			{
-				char *dn = ld->ld_options.ldo_cldapdn;
+				char *dn;
+				LDAP_MUTEX_LOCK( &ld->ld_options.ldo_mutex );
+				dn = ld->ld_options.ldo_cldapdn;
 				if (!dn) dn = "";
 				err = ber_printf( ber, "{isti",  /* '}' */
 					i, dn,
 					LDAP_REQ_ABANDON, msgid );
+				LDAP_MUTEX_UNLOCK( &ld->ld_options.ldo_mutex );
 			} else
 #endif
 			{
@@ -289,24 +278,32 @@ start_again:;
 	}
 
 	if ( lr != NULL ) {
+		LDAPConn *lc;
+		int freeconn = 0;
 		if ( sendabandon || lr->lr_status == LDAP_REQST_WRITING ) {
-			ldap_free_connection( ld, lr->lr_conn, 0, 1 );
+			freeconn = 1;
+			lc = lr->lr_conn;
 		}
-
 		if ( origid == msgid ) {
 			ldap_free_request( ld, lr );
 
 		} else {
 			lr->lr_abandoned = 1;
 		}
+
+		if ( freeconn ) {
+			/* release ld_req_mutex while grabbing ld_conn_mutex to
+			 * prevent deadlock.
+			 */
+			LDAP_MUTEX_UNLOCK( &ld->ld_req_mutex );
+			LDAP_MUTEX_LOCK( &ld->ld_conn_mutex );
+			ldap_free_connection( ld, lc, 0, 1 );
+			LDAP_MUTEX_UNLOCK( &ld->ld_conn_mutex );
+			LDAP_MUTEX_LOCK( &ld->ld_req_mutex );
+		}
 	}
 
-#ifdef LDAP_R_COMPILE
-	/* ld_abandoned is actually protected by the ld_res_mutex;
-	 * give up the ld_req_mutex and get the other */
-	ldap_pvt_thread_mutex_unlock( &ld->ld_req_mutex );
-	ldap_pvt_thread_mutex_lock( &ld->ld_res_mutex );
-#endif
+	LDAP_MUTEX_LOCK( &ld->ld_abandon_mutex );
 
 	/* use bisection */
 	i = 0;
@@ -320,10 +317,7 @@ start_again:;
 		ld->ld_errno = LDAP_SUCCESS;
 	}
 
-#ifdef LDAP_R_COMPILE
-	ldap_pvt_thread_mutex_unlock( &ld->ld_res_mutex );
-	ldap_pvt_thread_mutex_lock( &ld->ld_req_mutex );
-#endif
+	LDAP_MUTEX_UNLOCK( &ld->ld_abandon_mutex );
 	return( ld->ld_errno );
 }
 
@@ -348,14 +342,12 @@ ldap_int_bisect_find( ber_int_t *v, ber_len_t n, ber_int_t id, int *idxp )
 			end,
 			rc = 0;
 
-	assert( n >= 0 );
 	assert( id >= 0 );
 
 	begin = 0;
 	end = n - 1;
 
-	if ( n > 0 ) {
-		if ( id < v[ begin ] ) {
+		if ( n <= 0 || id < v[ begin ] ) {
 			*idxp = 0;
 
 		} else if ( id > v[ end ] ) {
@@ -365,7 +357,7 @@ ldap_int_bisect_find( ber_int_t *v, ber_len_t n, ber_int_t id, int *idxp )
 			int		pos;
 			ber_int_t	curid;
 	
-			while ( end >= begin ) {
+			do {
 				pos = (begin + end)/2;
 				curid = v[ pos ];
 	
@@ -373,25 +365,18 @@ ldap_int_bisect_find( ber_int_t *v, ber_len_t n, ber_int_t id, int *idxp )
 					end = pos - 1;
 	
 				} else if ( id > curid ) {
-					begin = pos + 1;
+					begin = ++pos;
 	
 				} else {
 					/* already abandoned? */
-					*idxp = pos;
 					rc = 1;
 					break;
 				}
-			}
+			} while ( end >= begin );
 	
-			if ( rc == 0 ) {
-				*idxp = pos + ( id > curid ? 1 : 0 );
-			}
+			*idxp = pos;
 		}
 
-	} else {
-		*idxp = 0;
-	}
-
 	return rc;
 }
 
@@ -417,9 +402,8 @@ ldap_int_bisect_insert( ber_int_t **vp, ber_len_t *np, int id, int idx )
 
 	assert( vp != NULL );
 	assert( np != NULL );
-	assert( *np >= 0 );
 	assert( idx >= 0 );
-	assert( idx <= *np );
+	assert( (unsigned) idx <= *np );
 
 	n = *np;
 
@@ -454,14 +438,12 @@ int
 ldap_int_bisect_delete( ber_int_t **vp, ber_len_t *np, int id, int idx )
 {
 	ber_int_t	*v;
-	ber_len_t	n;
-	int		i;
+	ber_len_t	i, n;
 
 	assert( vp != NULL );
 	assert( np != NULL );
-	assert( *np >= 0 );
 	assert( idx >= 0 );
-	assert( idx < *np );
+	assert( (unsigned) idx < *np );
 
 	v = *vp;
 
@@ -476,4 +458,3 @@ ldap_int_bisect_delete( ber_int_t **vp, ber_len_t *np, int id, int idx )
 
 	return 0;
 }
-
