@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/libraries/libldap/schema.c,v 1.77.2.1 2007/02/13 20:22:12 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2021 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,6 +30,8 @@
 
 #include <ldap_schema.h>
 
+static const char EndOfInput[] = "end of input";
+
 static const char *
 choose_name( char *names[], const char *fallback )
 {
@@ -39,48 +41,56 @@ choose_name( char *names[], const char *fallback )
 LDAP_CONST char *
 ldap_syntax2name( LDAPSyntax * syn )
 {
+	if (!syn) return NULL;
 	return( syn->syn_oid );
 }
 
 LDAP_CONST char *
 ldap_matchingrule2name( LDAPMatchingRule * mr )
 {
+	if (!mr) return NULL;
 	return( choose_name( mr->mr_names, mr->mr_oid ) );
 }
 
 LDAP_CONST char *
 ldap_matchingruleuse2name( LDAPMatchingRuleUse * mru )
 {
+	if (!mru) return NULL;
 	return( choose_name( mru->mru_names, mru->mru_oid ) );
 }
 
 LDAP_CONST char *
 ldap_attributetype2name( LDAPAttributeType * at )
 {
+	if (!at) return NULL;
 	return( choose_name( at->at_names, at->at_oid ) );
 }
 
 LDAP_CONST char *
 ldap_objectclass2name( LDAPObjectClass * oc )
 {
+	if (!oc) return NULL;
 	return( choose_name( oc->oc_names, oc->oc_oid ) );
 }
 
 LDAP_CONST char *
 ldap_contentrule2name( LDAPContentRule * cr )
 {
+	if (!cr) return NULL;
 	return( choose_name( cr->cr_names, cr->cr_oid ) );
 }
 
 LDAP_CONST char *
 ldap_nameform2name( LDAPNameForm * nf )
 {
+	if (!nf) return NULL;
 	return( choose_name( nf->nf_names, nf->nf_oid ) );
 }
 
 LDAP_CONST char *
 ldap_structurerule2name( LDAPStructureRule * sr )
 {
+	if (!sr) return NULL;
 	return( choose_name( sr->sr_names, NULL ) );
 }
 
@@ -377,7 +387,10 @@ struct berval *
 ldap_syntax2bv( LDAPSyntax * syn, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !syn || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -419,7 +432,10 @@ struct berval *
 ldap_matchingrule2bv( LDAPMatchingRule * mr, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !mr || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -478,7 +494,10 @@ struct berval *
 ldap_matchingruleuse2bv( LDAPMatchingRuleUse * mru, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !mru || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -537,7 +556,10 @@ struct berval *
 ldap_objectclass2bv( LDAPObjectClass * oc, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !oc || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -626,7 +648,10 @@ struct berval *
 ldap_contentrule2bv( LDAPContentRule * cr, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !cr || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -705,7 +730,10 @@ struct berval *
 ldap_structurerule2bv( LDAPStructureRule * sr, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !sr || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -769,7 +797,10 @@ struct berval *
 ldap_nameform2bv( LDAPNameForm * nf, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !nf || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -838,7 +869,10 @@ struct berval *
 ldap_attributetype2bv(  LDAPAttributeType * at, struct berval *bv )
 {
 	safe_string * ss;
-	
+
+	if ( !at || !bv )
+		return NULL;
+
 	ss = new_safe_string(256);
 	if ( !ss )
 		return NULL;
@@ -1307,7 +1341,7 @@ parse_oids(const char **sp, int *code, const int allow_quoted)
 
 	/*
 	 * Strictly speaking, doing this here accepts whsp before the
-	 * ( at the begining of an oidlist, but this is harmless.  Also,
+	 * ( at the beginning of an oidlist, but this is harmless.  Also,
 	 * we are very liberal in what we accept as an OID.  Maybe
 	 * refine later.
 	 */
@@ -1453,6 +1487,7 @@ free_extensions(LDAPSchemaExtensionItem **extensions)
 void
 ldap_syntax_free( LDAPSyntax * syn )
 {
+	if ( !syn ) return;
 	LDAP_FREE(syn->syn_oid);
 	if (syn->syn_names) LDAP_VFREE(syn->syn_names);
 	if (syn->syn_desc) LDAP_FREE(syn->syn_desc);
@@ -1514,7 +1549,7 @@ ldap_str2syntax( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_syntax_free(syn);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -1594,6 +1629,7 @@ ldap_str2syntax( LDAP_CONST char * s,
 void
 ldap_matchingrule_free( LDAPMatchingRule * mr )
 {
+	if (!mr) return;
 	LDAP_FREE(mr->mr_oid);
 	if (mr->mr_names) LDAP_VFREE(mr->mr_names);
 	if (mr->mr_desc) LDAP_FREE(mr->mr_desc);
@@ -1679,7 +1715,7 @@ ldap_str2matchingrule( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_matchingrule_free(mr);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -1793,6 +1829,7 @@ ldap_str2matchingrule( LDAP_CONST char * s,
 void
 ldap_matchingruleuse_free( LDAPMatchingRuleUse * mru )
 {
+	if (!mru) return;
 	LDAP_FREE(mru->mru_oid);
 	if (mru->mru_names) LDAP_VFREE(mru->mru_names);
 	if (mru->mru_desc) LDAP_FREE(mru->mru_desc);
@@ -1878,7 +1915,7 @@ ldap_str2matchingruleuse( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_matchingruleuse_free(mru);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -1991,6 +2028,7 @@ ldap_str2matchingruleuse( LDAP_CONST char * s,
 void
 ldap_attributetype_free(LDAPAttributeType * at)
 {
+	if (!at) return;
 	LDAP_FREE(at->at_oid);
 	if (at->at_names) LDAP_VFREE(at->at_names);
 	if (at->at_desc) LDAP_FREE(at->at_desc);
@@ -2088,6 +2126,11 @@ ldap_str2attributetype( LDAP_CONST char * s,
 					/* Non-numerical OID ... */
 					int len = ss-savepos;
 					at->at_oid = LDAP_MALLOC(len+1);
+					if ( !at->at_oid ) {
+						ldap_attributetype_free(at);
+						return NULL;
+					}
+
 					strncpy(at->at_oid, savepos, len);
 					at->at_oid[len] = 0;
 				}
@@ -2110,7 +2153,7 @@ ldap_str2attributetype( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_attributetype_free(at);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -2374,6 +2417,7 @@ ldap_str2attributetype( LDAP_CONST char * s,
 void
 ldap_objectclass_free(LDAPObjectClass * oc)
 {
+	if (!oc) return;
 	LDAP_FREE(oc->oc_oid);
 	if (oc->oc_names) LDAP_VFREE(oc->oc_names);
 	if (oc->oc_desc) LDAP_FREE(oc->oc_desc);
@@ -2460,6 +2504,11 @@ ldap_str2objectclass( LDAP_CONST char * s,
 					/* Non-numerical OID, ignore */
 					int len = ss-savepos;
 					oc->oc_oid = LDAP_MALLOC(len+1);
+					if ( !oc->oc_oid ) {
+						ldap_objectclass_free(oc);
+						return NULL;
+					}
+
 					strncpy(oc->oc_oid, savepos, len);
 					oc->oc_oid[len] = 0;
 				}
@@ -2483,7 +2532,7 @@ ldap_str2objectclass( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_objectclass_free(oc);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -2660,6 +2709,7 @@ ldap_str2objectclass( LDAP_CONST char * s,
 void
 ldap_contentrule_free(LDAPContentRule * cr)
 {
+	if (!cr) return;
 	LDAP_FREE(cr->cr_oid);
 	if (cr->cr_names) LDAP_VFREE(cr->cr_names);
 	if (cr->cr_desc) LDAP_FREE(cr->cr_desc);
@@ -2740,6 +2790,11 @@ ldap_str2contentrule( LDAP_CONST char * s,
 					/* Non-numerical OID, ignore */
 					int len = ss-savepos;
 					cr->cr_oid = LDAP_MALLOC(len+1);
+					if ( !cr->cr_oid ) {
+						ldap_contentrule_free(cr);
+						return NULL;
+					}
+
 					strncpy(cr->cr_oid, savepos, len);
 					cr->cr_oid[len] = 0;
 				}
@@ -2762,7 +2817,7 @@ ldap_str2contentrule( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_contentrule_free(cr);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -2917,6 +2972,7 @@ ldap_str2contentrule( LDAP_CONST char * s,
 void
 ldap_structurerule_free(LDAPStructureRule * sr)
 {
+	if (!sr) return;
 	if (sr->sr_names) LDAP_VFREE(sr->sr_names);
 	if (sr->sr_desc) LDAP_FREE(sr->sr_desc);
 	if (sr->sr_nameform) LDAP_FREE(sr->sr_nameform);
@@ -2987,7 +3043,7 @@ ldap_str2structurerule( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_structurerule_free(sr);
 			return NULL;
 		case TK_RIGHTPAREN:
@@ -3099,6 +3155,7 @@ ldap_str2structurerule( LDAP_CONST char * s,
 void
 ldap_nameform_free(LDAPNameForm * nf)
 {
+	if (!nf) return;
 	LDAP_FREE(nf->nf_oid);
 	if (nf->nf_names) LDAP_VFREE(nf->nf_names);
 	if (nf->nf_desc) LDAP_FREE(nf->nf_desc);
@@ -3176,7 +3233,7 @@ ldap_str2nameform( LDAP_CONST char * s,
 		switch (kind) {
 		case TK_EOS:
 			*code = LDAP_SCHERR_NORIGHTPAREN;
-			*errp = ss;
+			*errp = EndOfInput;
 			ldap_nameform_free(nf);
 			return NULL;
 		case TK_RIGHTPAREN:
