@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2020 The OpenLDAP Foundation.
+ * Copyright 2005-2021 The OpenLDAP Foundation.
  * Portions Copyright 2003 Howard Chu.
  * All rights reserved.
  *
@@ -33,7 +33,7 @@
 
 #include "back-ldap.h"
 
-#include "config.h"
+#include "slap-config.h"
 
 /*
  * From <draft-sermersheim-ldap-distproc>
