@@ -1,8 +1,8 @@
 /* init.c - initialize ldap backend */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldap/init.c,v 1.99.2.4 2007/02/13 20:22:24 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2003-2007 The OpenLDAP Foundation.
+ * Copyright 2003-2021 The OpenLDAP Foundation.
  * Portions Copyright 1999-2003 Howard Chu.
  * Portions Copyright 2000-2003 Pierangelo Masarati.
  * All rights reserved.
@@ -29,8 +29,21 @@
 #include <ac/socket.h>
 
 #include "slap.h"
+#include "slap-config.h"
 #include "back-ldap.h"
 
+static const ldap_extra_t ldap_extra = {
+	ldap_back_proxy_authz_ctrl,
+	ldap_back_controls_free,
+	slap_idassert_authzfrom_parse,
+	slap_idassert_passthru_parse_cf,
+	slap_idassert_parse,
+	slap_retry_info_destroy,
+	slap_retry_info_parse,
+	slap_retry_info_unparse,
+	ldap_back_connid2str
+};
+
 int
 ldap_back_open( BackendInfo	*bi )
 {
@@ -51,7 +64,10 @@ ldap_back_initialize( BackendInfo *bi )
 		 * and the entryTtl attribute */
 		SLAP_BFLAG_DYNAMIC |
 #endif /* LDAP_DYNAMIC_OBJECTS */
-		0;
+
+		/* back-ldap recognizes RFC4525 increment;
+		 * let the remote server complain, if needed (ITS#5912) */
+		SLAP_BFLAG_INCREMENT;
 
 	bi->bi_open = ldap_back_open;
 	bi->bi_config = 0;
@@ -82,23 +98,34 @@ ldap_back_initialize( BackendInfo *bi )
 	bi->bi_connection_init = 0;
 	bi->bi_connection_destroy = ldap_back_conn_destroy;
 
+	bi->bi_extra = (void *)&ldap_extra;
+
+	rc =  ldap_back_init_cf( bi );
+	if ( rc ) {
+		return rc;
+	}
+
 	rc = chain_initialize();
 	if ( rc ) {
 		return rc;
 	}
 
+	rc = pbind_initialize();
+	if ( rc ) {
+		return rc;
+	}
+
 #ifdef SLAP_DISTPROC
 	rc = distproc_initialize();
 	if ( rc ) {
 		return rc;
 	}
 #endif
-
-	return ldap_back_init_cf( bi );
+	return rc;
 }
 
 int
-ldap_back_db_init( Backend *be )
+ldap_back_db_init( Backend *be, ConfigReply *cr )
 {
 	ldapinfo_t	*li;
 	int		rc;
@@ -153,6 +180,11 @@ ldap_back_db_init( Backend *be )
 	}
 	li->li_conn_priv_max = LDAP_BACK_CONN_PRIV_DEFAULT;
 
+	ldap_pvt_thread_mutex_init( &li->li_counter_mutex );
+	for ( i = 0; i < SLAP_OP_LAST; i++ ) {
+		ldap_pvt_mp_init( li->li_ops_completed[ i ] );
+	}
+
 	be->be_private = li;
 	SLAP_DBFLAGS( be ) |= SLAP_DBFLAG_NOLASTMOD;
 
@@ -168,7 +200,7 @@ ldap_back_db_init( Backend *be )
 }
 
 int
-ldap_back_db_open( BackendDB *be )
+ldap_back_db_open( BackendDB *be, ConfigReply *cr )
 {
 	ldapinfo_t	*li = (ldapinfo_t *)be->be_private;
 
@@ -177,7 +209,7 @@ ldap_back_db_open( BackendDB *be )
 
 	Debug( LDAP_DEBUG_TRACE,
 		"ldap_back_db_open: URI=%s\n",
-		li->li_uri != NULL ? li->li_uri : "", 0, 0 );
+		li->li_uri != NULL ? li->li_uri : "" );
 
 	/* by default, use proxyAuthz control on each operation */
 	switch ( li->li_idassert_mode ) {
@@ -220,14 +252,10 @@ ldap_back_db_open( BackendDB *be )
 	if ( rc != 0 ) {
 		/* ignore by now */
 		rc = 0;
-#if 0
-		goto fail;
-#endif
 	}
 
 	li->li_flags |= LDAP_BACK_F_ISOPEN;
 
-fail:;
 	return rc;
 }
 
@@ -255,7 +283,7 @@ ldap_back_conn_free( void *v_lc )
 }
 
 int
-ldap_back_db_close( Backend *be )
+ldap_back_db_close( Backend *be, ConfigReply *cr )
 {
 	int		rc = 0;
 
@@ -267,7 +295,7 @@ ldap_back_db_close( Backend *be )
 }
 
 int
-ldap_back_db_destroy( Backend *be )
+ldap_back_db_destroy( Backend *be, ConfigReply *cr )
 {
 	if ( be->be_private ) {
 		ldapinfo_t	*li = ( ldapinfo_t * )be->be_private;
@@ -285,50 +313,11 @@ ldap_back_db_destroy( Backend *be )
 			ber_bvarray_free( li->li_bvuri );
 			li->li_bvuri = NULL;
 		}
-		if ( !BER_BVISNULL( &li->li_acl_authcID ) ) {
-			ch_free( li->li_acl_authcID.bv_val );
-			BER_BVZERO( &li->li_acl_authcID );
-		}
-		if ( !BER_BVISNULL( &li->li_acl_authcDN ) ) {
-			ch_free( li->li_acl_authcDN.bv_val );
-			BER_BVZERO( &li->li_acl_authcDN );
-		}
-		if ( !BER_BVISNULL( &li->li_acl_passwd ) ) {
-			ch_free( li->li_acl_passwd.bv_val );
-			BER_BVZERO( &li->li_acl_passwd );
-		}
-		if ( !BER_BVISNULL( &li->li_acl_sasl_mech ) ) {
-			ch_free( li->li_acl_sasl_mech.bv_val );
-			BER_BVZERO( &li->li_acl_sasl_mech );
-		}
-		if ( !BER_BVISNULL( &li->li_acl_sasl_realm ) ) {
-			ch_free( li->li_acl_sasl_realm.bv_val );
-			BER_BVZERO( &li->li_acl_sasl_realm );
-		}
-		if ( !BER_BVISNULL( &li->li_idassert_authcID ) ) {
-			ch_free( li->li_idassert_authcID.bv_val );
-			BER_BVZERO( &li->li_idassert_authcID );
-		}
-		if ( !BER_BVISNULL( &li->li_idassert_authcDN ) ) {
-			ch_free( li->li_idassert_authcDN.bv_val );
-			BER_BVZERO( &li->li_idassert_authcDN );
-		}
-		if ( !BER_BVISNULL( &li->li_idassert_passwd ) ) {
-			ch_free( li->li_idassert_passwd.bv_val );
-			BER_BVZERO( &li->li_idassert_passwd );
-		}
-		if ( !BER_BVISNULL( &li->li_idassert_authzID ) ) {
-			ch_free( li->li_idassert_authzID.bv_val );
-			BER_BVZERO( &li->li_idassert_authzID );
-		}
-		if ( !BER_BVISNULL( &li->li_idassert_sasl_mech ) ) {
-			ch_free( li->li_idassert_sasl_mech.bv_val );
-			BER_BVZERO( &li->li_idassert_sasl_mech );
-		}
-		if ( !BER_BVISNULL( &li->li_idassert_sasl_realm ) ) {
-			ch_free( li->li_idassert_sasl_realm.bv_val );
-			BER_BVZERO( &li->li_idassert_sasl_realm );
-		}
+
+		bindconf_free( &li->li_tls );
+		bindconf_free( &li->li_acl );
+		bindconf_free( &li->li_idassert.si_bc );
+
 		if ( li->li_idassert_authz != NULL ) {
 			ber_bvarray_free( li->li_idassert_authz );
 			li->li_idassert_authz = NULL;
@@ -352,6 +341,11 @@ ldap_back_db_destroy( Backend *be )
 		ldap_pvt_thread_mutex_unlock( &li->li_conninfo.lai_mutex );
 		ldap_pvt_thread_mutex_destroy( &li->li_conninfo.lai_mutex );
 		ldap_pvt_thread_mutex_destroy( &li->li_uri_mutex );
+
+		for ( i = 0; i < SLAP_OP_LAST; i++ ) {
+			ldap_pvt_mp_clear( li->li_ops_completed[ i ] );
+		}
+		ldap_pvt_thread_mutex_destroy( &li->li_counter_mutex );
 	}
 
 	ch_free( be->be_private );
