@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-perl/close.c,v 1.17.2.1 2007/02/13 20:22:27 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2021 The OpenLDAP Foundation.
  * Portions Copyright 1999 John C. Quillan.
  * Portions Copyright 2002 myinternet Limited.
  * All rights reserved.
@@ -16,7 +16,7 @@
  */
 
 #include "perl_back.h"
-
+#include "../slap-config.h"
 /**********************************************************
  *
  * Close
@@ -31,6 +31,9 @@ perl_back_close(
 	perl_destruct(PERL_INTERPRETER);
 	perl_free(PERL_INTERPRETER);
 	PERL_INTERPRETER = NULL;
+#ifdef PERL_SYS_TERM
+	PERL_SYS_TERM();
+#endif
 
 	ldap_pvt_thread_mutex_destroy( &perl_interpreter_mutex );	
 
@@ -39,9 +42,16 @@ perl_back_close(
 
 int
 perl_back_db_destroy(
-	BackendDB *be
+	BackendDB *be,
+	ConfigReply *cr
 )
 {
+	PerlBackend *pb = be->be_private;
+
+	ch_free( pb->pb_module_name );
+	ber_bvarray_free( pb->pb_module_path );
+	ber_bvarray_free( pb->pb_module_config );
+
 	free( be->be_private );
 	be->be_private = NULL;
 
