@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/entry-id.c,v 1.67.2.3 2007/02/13 20:22:28 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2021 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -36,7 +36,36 @@ struct berval backsql_baseObject_bv = BER_BVC( BACKSQL_BASEOBJECT_IDSTR );
 #endif /* BACKSQL_ARBITRARY_KEY */
 
 backsql_entryID *
-backsql_free_entryID( Operation *op, backsql_entryID *id, int freeit )
+backsql_entryID_dup( backsql_entryID *src, void *ctx )
+{
+	backsql_entryID	*dst;
+
+	if ( src == NULL ) return NULL;
+
+	dst = slap_sl_calloc( 1, sizeof( backsql_entryID ), ctx );
+	ber_dupbv_x( &dst->eid_ndn, &src->eid_ndn, ctx );
+	if ( src->eid_dn.bv_val == src->eid_ndn.bv_val ) {
+		dst->eid_dn = dst->eid_ndn;
+	} else {
+		ber_dupbv_x( &dst->eid_dn, &src->eid_dn, ctx );
+	}
+
+#ifdef BACKSQL_ARBITRARY_KEY
+	ber_dupbv_x( &dst->eid_id, &src->eid_id, ctx );
+	ber_dupbv_x( &dst->eid_keyval, &src->eid_keyval, ctx );
+#else /* ! BACKSQL_ARBITRARY_KEY */
+	dst->eid_id = src->eid_id;
+	dst->eid_keyval = src->eid_keyval;
+#endif /* ! BACKSQL_ARBITRARY_KEY */
+
+	dst->eid_oc = src->eid_oc;
+	dst->eid_oc_id = src->eid_oc_id;
+
+	return dst;
+}
+
+backsql_entryID *
+backsql_free_entryID( backsql_entryID *id, int freeit, void *ctx )
 {
 	backsql_entryID 	*next;
 
@@ -48,28 +77,28 @@ backsql_free_entryID( Operation *op, backsql_entryID *id, int freeit )
 		if ( !BER_BVISNULL( &id->eid_dn )
 				&& id->eid_dn.bv_val != id->eid_ndn.bv_val )
 		{
-			op->o_tmpfree( id->eid_dn.bv_val, op->o_tmpmemctx );
+			slap_sl_free( id->eid_dn.bv_val, ctx );
 			BER_BVZERO( &id->eid_dn );
 		}
 
-		op->o_tmpfree( id->eid_ndn.bv_val, op->o_tmpmemctx );
+		slap_sl_free( id->eid_ndn.bv_val, ctx );
 		BER_BVZERO( &id->eid_ndn );
 	}
 
 #ifdef BACKSQL_ARBITRARY_KEY
 	if ( !BER_BVISNULL( &id->eid_id ) ) {
-		op->o_tmpfree( id->eid_id.bv_val, op->o_tmpmemctx );
+		slap_sl_free( id->eid_id.bv_val, ctx );
 		BER_BVZERO( &id->eid_id );
 	}
 
 	if ( !BER_BVISNULL( &id->eid_keyval ) ) {
-		op->o_tmpfree( id->eid_keyval.bv_val, op->o_tmpmemctx );
+		slap_sl_free( id->eid_keyval.bv_val, ctx );
 		BER_BVZERO( &id->eid_keyval );
 	}
 #endif /* BACKSQL_ARBITRARY_KEY */
 
 	if ( freeit ) {
-		op->o_tmpfree( id, op->o_tmpmemctx );
+		slap_sl_free( id, ctx );
 	}
 
 	return next;
@@ -153,15 +182,15 @@ backsql_dn2id(
 	}
 	
 	/* begin TimesTen */
-	Debug( LDAP_DEBUG_TRACE, "   backsql_dn2id(\"%s\"): id_query \"%s\"\n",
-			ndn->bv_val, bi->sql_id_query, 0 );
 	assert( bi->sql_id_query != NULL );
+	Debug( LDAP_DEBUG_TRACE, "   backsql_dn2id(\"%s\"): id_query \"%s\"\n",
+			ndn->bv_val, bi->sql_id_query );
  	rc = backsql_Prepare( dbh, &sth, bi->sql_id_query, 0 );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, 
 			"   backsql_dn2id(\"%s\"): "
 			"error preparing SQL:\n   %s", 
-			ndn->bv_val, bi->sql_id_query, 0 );
+			ndn->bv_val, bi->sql_id_query );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		res = LDAP_OTHER;
 		goto done;
@@ -172,7 +201,7 @@ backsql_dn2id(
 		if ( backsql_api_dn2odbc( op, rs, &realndn ) ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_dn2id(\"%s\"): "
 				"backsql_api_dn2odbc(\"%s\") failed\n", 
-				ndn->bv_val, realndn.bv_val, 0 );
+				ndn->bv_val, realndn.bv_val );
 			res = LDAP_OTHER;
 			goto done;
 		}
@@ -193,7 +222,7 @@ backsql_dn2id(
 
 		Debug( LDAP_DEBUG_TRACE, "   backsql_dn2id(\"%s\"): "
 				"upperdn=\"%s\"\n",
-				ndn->bv_val, upperdn, 0 );
+				ndn->bv_val, upperdn );
 		ber_str2bv( upperdn, 0, 0, &tbbDN );
 
 	} else {
@@ -203,7 +232,7 @@ backsql_dn2id(
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_dn2id(\"%s\"): "
 				"upperdn=\"%s\"\n",
-				ndn->bv_val, upperdn, 0 );
+				ndn->bv_val, upperdn );
 			ber_str2bv( upperdn, 0, 0, &tbbDN );
 
 		} else {
@@ -216,7 +245,7 @@ backsql_dn2id(
 		/* end TimesTen */ 
 		Debug( LDAP_DEBUG_TRACE, "   backsql_dn2id(\"%s\"): "
 			"error binding dn=\"%s\" parameter:\n", 
-			ndn->bv_val, tbbDN.bv_val, 0 );
+			ndn->bv_val, tbbDN.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		res = LDAP_OTHER;
 		goto done;
@@ -235,16 +264,11 @@ backsql_dn2id(
 	backsql_BindRowAsStrings_x( sth, &row, op->o_tmpmemctx );
 	rc = SQLFetch( sth );
 	if ( BACKSQL_SUCCESS( rc ) ) {
-		char	buf[ SLAP_TEXT_BUFLEN ];
-
 #ifdef LDAP_DEBUG
-		snprintf( buf, sizeof(buf),
-			"id=%s keyval=%s oc_id=%s dn=%s",
-			row.cols[ 0 ], row.cols[ 1 ],
-			row.cols[ 2 ], row.cols[ 3 ] );
-		Debug( LDAP_DEBUG_TRACE,
-			"   backsql_dn2id(\"%s\"): %s\n",
-			ndn->bv_val, buf, 0 );
+		Debug(LDAP_DEBUG_TRACE,
+		      "   backsql_dn2id(\"%s\"): id=%s keyval=%s oc_id=%s dn=%s\n",
+		      ndn->bv_val, row.cols[0], row.cols[1], row.cols[2],
+		      row.cols[3] );
 #endif /* LDAP_DEBUG */
 
 		res = LDAP_SUCCESS;
@@ -259,16 +283,16 @@ backsql_dn2id(
 			ber_str2bv_x( row.cols[ 1 ], 0, 1, &id->eid_keyval,
 					op->o_tmpmemctx );
 #else /* ! BACKSQL_ARBITRARY_KEY */
-			if ( lutil_atoulx( &id->eid_id, row.cols[ 0 ], 0 ) != 0 ) {
+			if ( BACKSQL_STR2ID( &id->eid_id, row.cols[ 0 ], 0 ) != 0 ) {
 				res = LDAP_OTHER;
 				goto done;
 			}
-			if ( lutil_atoulx( &id->eid_keyval, row.cols[ 1 ], 0 ) != 0 ) {
+			if ( BACKSQL_STR2ID( &id->eid_keyval, row.cols[ 1 ], 0 ) != 0 ) {
 				res = LDAP_OTHER;
 				goto done;
 			}
 #endif /* ! BACKSQL_ARBITRARY_KEY */
-			if ( lutil_atoulx( &id->eid_oc_id, row.cols[ 2 ], 0 ) != 0 ) {
+			if ( BACKSQL_STR2ID( &id->eid_oc_id, row.cols[ 2 ], 0 ) != 0 ) {
 				res = LDAP_OTHER;
 				goto done;
 			}
@@ -291,7 +315,7 @@ backsql_dn2id(
 					ldap_err2string( res ) );
 
 				/* cleanup... */
-				(void)backsql_free_entryID( op, id, 0 );
+				(void)backsql_free_entryID( id, 0, op->o_tmpmemctx );
 			}
 
 			if ( dn.bv_val != row.cols[ 3 ] ) {
@@ -340,7 +364,7 @@ done:;
 
 	Debug( LDAP_DEBUG_TRACE,
 		"<==backsql_dn2id(\"%s\"): err=%d\n",
-		ndn->bv_val, res, 0 );
+		ndn->bv_val, res );
 	if ( sth != SQL_NULL_HSTMT ) {
 		SQLFreeStmt( sth, SQL_DROP );
 	}
@@ -366,7 +390,7 @@ backsql_count_children(
 	int			res = LDAP_SUCCESS;
 
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_count_children(): dn=\"%s\"\n", 
-			dn->bv_val, 0, 0 );
+			dn->bv_val );
 
 	if ( dn->bv_len > BACKSQL_MAX_DN_LEN ) {
 		Debug( LDAP_DEBUG_TRACE, 
@@ -377,14 +401,14 @@ backsql_count_children(
 	}
 	
 	/* begin TimesTen */
-	Debug(LDAP_DEBUG_TRACE, "children id query \"%s\"\n", 
-			bi->sql_has_children_query, 0, 0);
 	assert( bi->sql_has_children_query != NULL );
+	Debug(LDAP_DEBUG_TRACE, "children id query \"%s\"\n", 
+			bi->sql_has_children_query );
  	rc = backsql_Prepare( dbh, &sth, bi->sql_has_children_query, 0 );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, 
 			"backsql_count_children(): error preparing SQL:\n%s", 
-			bi->sql_has_children_query, 0, 0);
+			bi->sql_has_children_query );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		return LDAP_OTHER;
@@ -395,7 +419,7 @@ backsql_count_children(
 		/* end TimesTen */ 
 		Debug( LDAP_DEBUG_TRACE, "backsql_count_children(): "
 			"error binding dn=\"%s\" parameter:\n", 
-			dn->bv_val, 0, 0 );
+			dn->bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		return LDAP_OTHER;
@@ -405,7 +429,7 @@ backsql_count_children(
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_count_children(): "
 			"error executing query (\"%s\", \"%s\"):\n", 
-			bi->sql_has_children_query, dn->bv_val, 0 );
+			bi->sql_has_children_query, dn->bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		return LDAP_OTHER;
@@ -450,7 +474,7 @@ backsql_count_children(
 	SQLFreeStmt( sth, SQL_DROP );
 
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_count_children(): %lu\n",
-			*nchildren, 0, 0 );
+			*nchildren );
 
 	return res;
 }
@@ -478,7 +502,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 {
 	backsql_at_map_rec	*at = v_at;
 	backsql_srch_info	*bsi = v_bsi;
-	backsql_info		*bi = (backsql_info *)bsi->bsi_op->o_bd->be_private;
+	backsql_info		*bi;
 	RETCODE			rc;
 	SQLHSTMT		sth = SQL_NULL_HSTMT;
 	BACKSQL_ROW_NTS		row;
@@ -502,22 +526,16 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 
 	assert( at != NULL );
 	assert( bsi != NULL );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_get_attr_vals(): "
+		"oc=\"%s\" attr=\"%s\" keyval=" BACKSQL_IDFMT "\n",
+		BACKSQL_OC_NAME( bsi->bsi_oc ), at->bam_ad->ad_cname.bv_val, 
+		BACKSQL_IDARG(bsi->bsi_c_eid->eid_keyval) );
 
-#ifdef BACKSQL_ARBITRARY_KEY
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_get_attr_vals(): "
-		"oc=\"%s\" attr=\"%s\" keyval=%s\n",
-		BACKSQL_OC_NAME( bsi->bsi_oc ), at->bam_ad->ad_cname.bv_val, 
-		bsi->bsi_c_eid->eid_keyval.bv_val );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_get_attr_vals(): "
-		"oc=\"%s\" attr=\"%s\" keyval=%ld\n",
-		BACKSQL_OC_NAME( bsi->bsi_oc ), at->bam_ad->ad_cname.bv_val, 
-		bsi->bsi_c_eid->eid_keyval );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+	bi = (backsql_info *)bsi->bsi_op->o_bd->be_private;
 
 #ifdef BACKSQL_PRETTY_VALIDATE
-	validate = at->bam_ad->ad_type->sat_syntax->ssyn_validate;
-	pretty =  at->bam_ad->ad_type->sat_syntax->ssyn_pretty;
+	validate = at->bam_true_ad->ad_type->sat_syntax->ssyn_validate;
+	pretty =  at->bam_true_ad->ad_type->sat_syntax->ssyn_pretty;
 
 	if ( validate == NULL && pretty == NULL ) {
 		return 1;
@@ -525,8 +543,8 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 #endif /* BACKSQL_PRETTY_VALIDATE */
 
 #ifdef BACKSQL_COUNTQUERY
-	if ( at->bam_ad->ad_type->sat_equality ) {
-		normfunc = at->bam_ad->ad_type->sat_equality->smr_normalize;
+	if ( at->bam_true_ad->ad_type->sat_equality ) {
+		normfunc = at->bam_true_ad->ad_type->sat_equality->smr_normalize;
 	}
 
 	/* Count how many rows will be returned. This avoids memory 
@@ -537,7 +555,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
 			"error preparing count query: %s\n",
-			at->bam_countquery, 0, 0 );
+			at->bam_countquery );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 		return 1;
 	}
@@ -546,7 +564,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 			&bsi->bsi_c_eid->eid_keyval );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
-			"error binding key value parameter\n", 0, 0, 0 );
+			"error binding key value parameter\n" );
 		SQLFreeStmt( sth, SQL_DROP );
 		return 1;
 	}
@@ -555,7 +573,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 	if ( ! BACKSQL_SUCCESS( rc ) ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
 			"error executing attribute count query '%s'\n",
-			at->bam_countquery, 0, 0 );
+			at->bam_countquery );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		return 1;
@@ -570,26 +588,25 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
 			"error fetch results of count query: %s\n",
-			at->bam_countquery, 0, 0 );
+			at->bam_countquery );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		return 1;
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
-		"number of values in query: %u\n", count, 0, 0 );
+		"number of values in query: %u\n", count );
 	SQLFreeStmt( sth, SQL_DROP );
 	if ( count == 0 ) {
 		return 1;
 	}
 
-	attr = attr_find( bsi->bsi_e->e_attrs, at->bam_ad );
+	attr = attr_find( bsi->bsi_e->e_attrs, at->bam_true_ad );
 	if ( attr != NULL ) {
 		BerVarray	tmp;
 
 		if ( attr->a_vals != NULL ) {
-			for ( ; !BER_BVISNULL( &attr->a_vals[ oldcount ] ); oldcount++ )
-				/* just count */ ;
+			oldcount = attr->a_numvals;
 		}
 
 		tmp = ch_realloc( attr->a_vals, ( oldcount + count + 1 ) * sizeof( struct berval ) );
@@ -610,19 +627,20 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 		} else {
 			attr->a_nvals = attr->a_vals;
 		}
+		attr->a_numvals += count;
 
 	} else {
 		append = 1;
 
 		/* Make space for the array of values */
-		attr = attr_alloc( at->bam_ad );
+		attr = attr_alloc( at->bam_true_ad );
+		attr->a_numvals = count;
 		attr->a_vals = ch_calloc( count + 1, sizeof( struct berval ) );
 		if ( attr->a_vals == NULL ) {
-			Debug( LDAP_DEBUG_TRACE, "Out of memory!\n", 0,0,0 );
+			Debug( LDAP_DEBUG_TRACE, "Out of memory!\n" );
 			ch_free( attr );
 			return 1;
 		}
-		memset( attr->a_vals, 0, ( count + 1 ) * sizeof( struct berval ) );
 		if ( normfunc ) {
 			attr->a_nvals = ch_calloc( count + 1, sizeof( struct berval ) );
 			if ( attr->a_nvals == NULL ) {
@@ -630,8 +648,6 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 				ch_free( attr );
 				return 1;
 
-			} else {
-				memset( attr->a_nvals, 0, ( count + 1 ) * sizeof( struct berval ) );
 			}
 
 		} else {
@@ -643,7 +659,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 	rc = backsql_Prepare( bsi->bsi_dbh, &sth, at->bam_query, 0 );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
-			"error preparing query: %s\n", at->bam_query, 0, 0 );
+			"error preparing query: %s\n", at->bam_query );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 #ifdef BACKSQL_COUNTQUERY
 		if ( append ) {
@@ -657,7 +673,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 			&bsi->bsi_c_eid->eid_keyval );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
-			"error binding key value parameter\n", 0, 0, 0 );
+			"error binding key value parameter\n" );
 #ifdef BACKSQL_COUNTQUERY
 		if ( append ) {
 			attr_free( attr );
@@ -667,22 +683,16 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 	}
 
 #ifdef BACKSQL_TRACE
-#ifdef BACKSQL_ARBITRARY_KEY
 	Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
-		"query=\"%s\" keyval=%s\n", at->bam_query,
-		bsi->bsi_c_eid->eid_keyval.bv_val, 0 );
-#else /* !BACKSQL_ARBITRARY_KEY */
-	Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
-		"query=\"%s\" keyval=%d\n", at->bam_query,
-		bsi->bsi_c_eid->eid_keyval, 0 );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+		"query=\"%s\" keyval=" BACKSQL_IDFMT "\n", at->bam_query,
+		BACKSQL_IDARG(bsi->bsi_c_eid->eid_keyval) );
 #endif /* BACKSQL_TRACE */
 
 	rc = SQLExecute( sth );
 	if ( ! BACKSQL_SUCCESS( rc ) ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_get_attr_vals(): "
 			"error executing attribute query \"%s\"\n",
-			at->bam_query, 0, 0 );
+			at->bam_query );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 #ifdef BACKSQL_COUNTQUERY
@@ -735,48 +745,59 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 				}
 #endif /* BACKSQL_TRACE */
 
-				/*
-				 * FIXME: what if a binary 
-				 * is fetched?
+				/* ITS#3386, ITS#3113 - 20070308
+				 * If a binary is fetched?
+				 * must use the actual size read
+				 * from the database.
 				 */
-				ber_str2bv( row.cols[ i ], 0, 0, &bv );
+				if ( BACKSQL_IS_BINARY( row.col_type[ i ] ) ) {
+#ifdef BACKSQL_TRACE
+					Debug( LDAP_DEBUG_ANY,
+						"==>backsql_get_attr_vals(\"%s\"): "
+						"column name %s: data is binary; "
+						"using database size %ld\n",
+						bsi->bsi_e->e_name.bv_val,
+						ad->ad_cname.bv_val,
+						row.value_len[ i ] );
+#endif /* BACKSQL_TRACE */
+					bv.bv_val = row.cols[ i ];
+					bv.bv_len = row.value_len[ i ];
+
+				} else {
+					ber_str2bv( row.cols[ i ], 0, 0, &bv );
+				}
 
 #ifdef BACKSQL_PRETTY_VALIDATE
 				if ( pretty ) {
 					struct berval	pbv;
 
-					retval = pretty( at->bam_ad->ad_type->sat_syntax,
+					retval = pretty( at->bam_true_ad->ad_type->sat_syntax,
 						&bv, &pbv, bsi->bsi_op->o_tmpmemctx );
 					bv = pbv;
 
 				} else {
-					retval = validate( at->bam_ad->ad_type->sat_syntax,
+					retval = validate( at->bam_true_ad->ad_type->sat_syntax,
 						&bv );
 				}
 
 				if ( retval != LDAP_SUCCESS ) {
-					char	buf[ SLAP_TEXT_BUFLEN ];
-
 					/* FIXME: we're ignoring invalid values,
 					 * but we're accepting the attributes;
 					 * should we fail at all? */
-					snprintf( buf, sizeof( buf ),
-							"unable to %s value #%lu "
-							"of AttributeDescription %s",
-							pretty ? "prettify" : "validate",
-							k - oldcount,
-							at->bam_ad->ad_cname.bv_val );
-					Debug( LDAP_DEBUG_TRACE,
-						"==>backsql_get_attr_vals(\"%s\"): "
-						"%s (%d)\n",
-						bsi->bsi_e->e_name.bv_val, buf, retval );
+					Debug(LDAP_DEBUG_TRACE,
+					      "==>backsql_get_attr_vals(\"%s\"): " "unable to %s value #%lu " "of AttributeDescription %s (%d)\n",
+					      bsi->bsi_e->e_name.bv_val,
+					      pretty ? "prettify" : "validate",
+					      k - oldcount,
+					      at->bam_ad->ad_cname.bv_val,
+					      retval );
 					continue;
 				}
 #endif /* BACKSQL_PRETTY_VALIDATE */
 
 #ifndef BACKSQL_COUNTQUERY
 				(void)backsql_entry_addattr( bsi->bsi_e, 
-						at->bam_ad, &bv,
+						at->bam_true_ad, &bv,
 						bsi->bsi_op->o_tmpmemctx );
 
 #else /* BACKSQL_COUNTQUERY */
@@ -784,26 +805,21 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 					struct berval	nbv;
 
 					retval = (*normfunc)( SLAP_MR_VALUE_OF_ATTRIBUTE_SYNTAX,
-						at->bam_ad->ad_type->sat_syntax,
-						at->bam_ad->ad_type->sat_equality,
+						at->bam_true_ad->ad_type->sat_syntax,
+						at->bam_true_ad->ad_type->sat_equality,
 						&bv, &nbv,
 						bsi->bsi_op->o_tmpmemctx );
 
 					if ( retval != LDAP_SUCCESS ) {
-						char	buf[ SLAP_TEXT_BUFLEN ];
-
 						/* FIXME: we're ignoring invalid values,
 						 * but we're accepting the attributes;
 						 * should we fail at all? */
-						snprintf( buf, sizeof( buf ),
-							"unable to normalize value #%lu "
-							"of AttributeDescription %s",
-							k - oldcount,
-							at->bam_ad->ad_cname.bv_val );
-						Debug( LDAP_DEBUG_TRACE,
-							"==>backsql_get_attr_vals(\"%s\"): "
-							"%s (%d)\n",
-							bsi->bsi_e->e_name.bv_val, buf, retval );
+						Debug(LDAP_DEBUG_TRACE,
+						      "==>backsql_get_attr_vals(\"%s\"): " "unable to normalize value #%lu " "of AttributeDescription %s (%d)\n",
+						      bsi->bsi_e->e_name.bv_val,
+						      k - oldcount,
+						      at->bam_ad->ad_cname.bv_val,
+						      retval );
 
 #ifdef BACKSQL_PRETTY_VALIDATE
 						if ( pretty ) {
@@ -834,12 +850,12 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 
 #ifdef BACKSQL_TRACE
 				Debug( LDAP_DEBUG_TRACE, "prec=%d\n",
-					(int)row.col_prec[ i ], 0, 0 );
+					(int)row.col_prec[ i ] );
 
 			} else {
       				Debug( LDAP_DEBUG_TRACE, "NULL value "
 					"in this row for attribute \"%s\"\n",
-					row.col_names[ i ].bv_val, 0, 0 );
+					row.col_names[ i ].bv_val );
 #endif /* BACKSQL_TRACE */
 			}
 		}
@@ -860,7 +876,7 @@ backsql_get_attr_vals( void *v_at, void *v_bsi )
 #endif /* BACKSQL_COUNTQUERY */
 
 	SQLFreeStmt( sth, SQL_DROP );
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_get_attr_vals()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_get_attr_vals()\n" );
 
 	if ( at->bam_next ) {
 		res = backsql_get_attr_vals( at->bam_next, v_bsi );
@@ -884,35 +900,37 @@ backsql_id2entry( backsql_srch_info *bsi, backsql_entryID *eid )
 	int			i;
 	int			rc;
 
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_id2entry()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_id2entry()\n" );
 
 	assert( bsi->bsi_e != NULL );
 
 	memset( bsi->bsi_e, 0, sizeof( Entry ) );
 
 	if ( bi->sql_baseObject && BACKSQL_IS_BASEOBJECT_ID( &eid->eid_id ) ) {
-		Entry	*e;
-
-		e = entry_dup( bi->sql_baseObject );
-		if ( e == NULL ) {
-			return LDAP_NO_MEMORY;
-		}
-			
-		*bsi->bsi_e = *e;
-		free( e );
+		(void)entry_dup2( bsi->bsi_e, bi->sql_baseObject );
 		goto done;
 	}
 
-	ber_dupbv_x( &bsi->bsi_e->e_name, &eid->eid_dn, op->o_tmpmemctx );
-	ber_dupbv_x( &bsi->bsi_e->e_nname, &eid->eid_ndn, op->o_tmpmemctx );
-
 	bsi->bsi_e->e_attrs = NULL;
 	bsi->bsi_e->e_private = NULL;
 
-	bsi->bsi_oc = backsql_id2oc( bsi->bsi_op->o_bd->be_private,
+	if ( eid->eid_oc == NULL ) {
+		eid->eid_oc = backsql_id2oc( bsi->bsi_op->o_bd->be_private,
 			eid->eid_oc_id );
+		if ( eid->eid_oc == NULL ) {
+			Debug( LDAP_DEBUG_TRACE,
+				"backsql_id2entry(): unable to fetch objectClass with id=" BACKSQL_IDNUMFMT " for entry id=" BACKSQL_IDFMT " dn=\"%s\"\n",
+				eid->eid_oc_id, BACKSQL_IDARG(eid->eid_id),
+				eid->eid_dn.bv_val );
+			return LDAP_OTHER;
+		}
+	}
+	bsi->bsi_oc = eid->eid_oc;
 	bsi->bsi_c_eid = eid;
 
+	ber_dupbv_x( &bsi->bsi_e->e_name, &eid->eid_dn, op->o_tmpmemctx );
+	ber_dupbv_x( &bsi->bsi_e->e_nname, &eid->eid_ndn, op->o_tmpmemctx );
+
 #ifndef BACKSQL_ARBITRARY_KEY	
 	/* FIXME: unused */
 	bsi->bsi_e->e_id = eid->eid_id;
@@ -930,13 +948,13 @@ backsql_id2entry( backsql_srch_info *bsi, backsql_entryID *eid )
 	if ( bsi->bsi_attrs == NULL || ( bsi->bsi_flags & BSQL_SF_ALL_USER ) )
 	{
 		Debug( LDAP_DEBUG_TRACE, "backsql_id2entry(): "
-			"retrieving all attributes\n", 0, 0, 0 );
+			"retrieving all attributes\n" );
 		avl_apply( bsi->bsi_oc->bom_attrs, backsql_get_attr_vals,
 				bsi, 0, AVL_INORDER );
 
 	} else {
 		Debug( LDAP_DEBUG_TRACE, "backsql_id2entry(): "
-			"custom attribute list\n", 0, 0, 0 );
+			"custom attribute list\n" );
 		for ( i = 0; !BER_BVISNULL( &bsi->bsi_attrs[ i ].an_name ); i++ ) {
 			backsql_at_map_rec	**vat;
 			AttributeName		*an = &bsi->bsi_attrs[ i ];
@@ -965,9 +983,9 @@ backsql_id2entry( backsql_srch_info *bsi, backsql_entryID *eid )
 			if ( rc != 0 || vat == NULL ) {
 				Debug( LDAP_DEBUG_TRACE, "backsql_id2entry(): "
 						"attribute \"%s\" is not defined "
-						"for objectlass \"%s\"\n",
+						"for objectclass \"%s\"\n",
 						an->an_name.bv_val, 
-						BACKSQL_OC_NAME( bsi->bsi_oc ), 0 );
+						BACKSQL_OC_NAME( bsi->bsi_oc ) );
 				continue;
 			}
 
@@ -996,15 +1014,16 @@ next:;
 	}
 
 	if ( ( bsi->bsi_flags & BSQL_SF_ALL_OPER )
-			|| an_find( bsi->bsi_attrs, &AllOper )
+			|| an_find( bsi->bsi_attrs, slap_bv_all_operational_attrs )
 			|| an_find( bsi->bsi_attrs, &slap_schema.si_ad_structuralObjectClass->ad_cname ) )
 	{
+		ObjectClass	*soc = NULL;
+
 		if ( BACKSQL_CHECK_SCHEMA( bi ) ) {
 			Attribute	*a;
 			const char	*text = NULL;
 			char		textbuf[ 1024 ];
 			size_t		textlen = sizeof( textbuf );
-			ObjectClass	*soc = NULL;
 			struct berval	bv[ 2 ],
 					*nvals;
 			int		rc = LDAP_SUCCESS;
@@ -1032,20 +1051,32 @@ next:;
 			}
 
 			if ( !bvmatch( &soc->soc_cname, &bsi->bsi_oc->bom_oc->soc_cname ) ) {
+				if ( !is_object_subclass( bsi->bsi_oc->bom_oc, soc ) ) {
+      					Debug( LDAP_DEBUG_TRACE, "backsql_id2entry(%s): "
+						"computed structuralObjectClass %s "
+						"does not match objectClass %s associated "
+						"to entry\n",
+						bsi->bsi_e->e_name.bv_val, soc->soc_cname.bv_val,
+						bsi->bsi_oc->bom_oc->soc_cname.bv_val );
+					backsql_entry_clean( op, bsi->bsi_e );
+					return rc;
+				}
+
       				Debug( LDAP_DEBUG_TRACE, "backsql_id2entry(%s): "
 					"computed structuralObjectClass %s "
-					"does not match objectClass %s associated "
+					"is subclass of objectClass %s associated "
 					"to entry\n",
 					bsi->bsi_e->e_name.bv_val, soc->soc_cname.bv_val,
 					bsi->bsi_oc->bom_oc->soc_cname.bv_val );
-				backsql_entry_clean( op, bsi->bsi_e );
-				return rc;
 			}
+
+		} else {
+			soc = bsi->bsi_oc->bom_oc;
 		}
 
 		rc = attr_merge_normalize_one( bsi->bsi_e,
 				slap_schema.si_ad_structuralObjectClass,
-				&bsi->bsi_oc->bom_oc->soc_cname,
+				&soc->soc_cname,
 				bsi->bsi_op->o_tmpmemctx );
 		if ( rc != LDAP_SUCCESS ) {
 			backsql_entry_clean( op, bsi->bsi_e );
@@ -1054,7 +1085,7 @@ next:;
 	}
 
 done:;
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_id2entry()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_id2entry()\n" );
 
 	return LDAP_SUCCESS;
 }
