@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2020 The OpenLDAP Foundation.
+ * Copyright 1998-2021 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -210,7 +210,7 @@ int connections_shutdown(void)
  */
 int connections_timeout_idle(time_t now)
 {
-	int i = 0, writers = 0;
+	int i = 0;
 	ber_socket_t connindex;
 	Connection* c;
 
@@ -520,7 +520,12 @@ Connection * connection_init(
 		BER_BVZERO( &c->c_peer_name );
 
 		ber_sockbuf_free( c->c_sb );
-		c->c_sb = NULL;
+		c->c_sb = ber_sockbuf_alloc( );
+		{
+			ber_len_t max = sockbuf_max_incoming;
+			ber_sockbuf_ctrl( c->c_sb, LBER_SB_OPT_SET_MAX_INCOMING, &max );
+		}
+
 		c->c_sd = AC_SOCKET_INVALID;
 		ldap_pvt_thread_mutex_unlock( &c->c_mutex );
 
@@ -2018,7 +2023,7 @@ int connection_write(ber_socket_t s)
 	return 0;
 }
 
-int connection_write_resume( Connection *c )
+void connection_write_resume( Connection *c )
 {
 	Operation *op;
 
