@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2020 The OpenLDAP Foundation.
+ * Copyright 2005-2021 The OpenLDAP Foundation.
  * Portions Copyright 2005-2006 SysNet s.n.c.
  * All rights reserved.
  *
@@ -31,7 +31,7 @@
 #include "lutil.h"
 #include "ldap_rq.h"
 
-#include "config.h"
+#include "slap-config.h"
 
 #define	DDS_RF2589_MAX_TTL		(31557600)	/* 1 year + 6 hours */
 #define	DDS_RF2589_DEFAULT_TTL		(86400)		/* 1 day */
@@ -155,6 +155,7 @@ dds_expire( void *ctx, dds_info_t *di )
 	op->ors_tlimit = DDS_INTERVAL( di )/2 + 1;
 	op->ors_slimit = SLAP_NO_LIMIT;
 	op->ors_attrs = slap_anlist_no_attrs;
+	op->o_do_not_cache = 1;
 
 	expire = slap_get_time() - di->di_tolerance;
 	ts.bv_val = tsbuf;
@@ -1722,6 +1723,7 @@ dds_count( void *ctx, BackendDB *be )
 	op->ors_tlimit = SLAP_NO_LIMIT;
 	op->ors_slimit = SLAP_NO_LIMIT;
 	op->ors_attrs = slap_anlist_no_attrs;
+	op->o_do_not_cache = 1;
 
 	op->ors_filterstr.bv_len = STRLENOF( "(objectClass=" ")" )
 		+ slap_schema.si_oc_dynamicObject->soc_cname.bv_len;
