@@ -1,7 +1,7 @@
 /* seqmod.c - sequenced modifies */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2004-2020 The OpenLDAP Foundation.
+ * Copyright 2004-2021 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -22,7 +22,7 @@
 #ifdef SLAPD_OVER_SEQMOD
 
 #include "slap.h"
-#include "config.h"
+#include "slap-config.h"
 
 /* This overlay serializes concurrent attempts to modify a single entry */
 
