@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2020 The OpenLDAP Foundation.
+ * Copyright 1998-2021 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -31,8 +31,8 @@
 
 LDAP_BEGIN_DECL
 
-struct config_args_s;	/* config.h */
-struct config_reply_s;	/* config.h */
+struct config_args_s;	/* slap-config.h */
+struct config_reply_s;	/* slap-config.h */
 
 /*
  * aci.c
@@ -809,7 +809,7 @@ LDAP_SLAPD_F (const char *) connection_state2str LDAP_P(( int state ))
 
 LDAP_SLAPD_F (int) connection_read_activate LDAP_P((ber_socket_t s));
 LDAP_SLAPD_F (int) connection_write LDAP_P((ber_socket_t s));
-LDAP_SLAPD_F (int) connection_write_resume LDAP_P((Connection *c));
+LDAP_SLAPD_F (void) connection_write_resume LDAP_P((Connection *c));
 
 LDAP_SLAPD_F (void) connection_op_finish LDAP_P((
 	Operation *op ));
