@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2002-2020 The OpenLDAP Foundation.
+ * Copyright 2002-2021 The OpenLDAP Foundation.
  * Portions Copyright 1997,2002-2003 IBM Corporation.
  * All rights reserved.
  *
@@ -23,7 +23,7 @@
 #include "portable.h"
 #include "ldap_pvt_thread.h"
 #include "slap.h"
-#include "config.h"
+#include "slap-config.h"
 #include "slapi.h"
 #include "lutil.h"
 
